/*******************************************************************************
*  PPM_decoder.c
*
*   PPM:  
*
*       Copyright (c) 2009 Nick Shl
*           All rights reserved.
*
*
*  :
*
*  Aug 25, 2009  Nick_Shl   
*
*  Chip type           : ATtiny2313
*  Clock frequency     : 8,000000 MHz
*  Memory model        : Tiny
*  External SRAM size  : 0
*  Data Stack size     : 16
*/// ***************************************************************************

// *****************************************************************************
// ***       ***************************************************
// *****************************************************************************
#include <tiny2313.h>
#include <Delay.h>

// *****************************************************************************
// ***      *********************************************************
// *****************************************************************************

//   
#define AVR_Clock_Freq 8000000
//  
#define TimerDevider 64
//     1 
#define TimerClockPerSec(s) (int)((s * AVR_Clock_Freq) / TimerDevider)
//    0  1 
#define Timer0_Ovf_Per_Sec(s) (int)((s * AVR_Clock_Freq) / TimerDevider / 0xFF))

//   PPM -   8.    
// ,   -    
#define ChannelsNum 8

//    
#define RetryCount 16

//  () 
#define OUTPORT PORTB
//  ()   0  .
//  2 ()    PORT.2,   PORT.3  ..
#define FCShift 2

//  PPM 
#define INPUT PIND.2

//   
#define RedLED(x)   PORTD.4 = x
#define GreenLED(x) PORTD.5 = x

//    ,   
#define True  1
#define False 0
#define On    1
#define Off   0

// *****************************************************************************
// ***       ***********************************************
// *****************************************************************************

//   
register unsigned char i;

//    
register unsigned char OUT_CH_NUM;
//     -    
register unsigned int ChannelsTime = 0;
//   .      PPM .
unsigned char SendPacketCount;

//     -     16 
volatile unsigned char Timer0_ovf;
//  PPM
unsigned char Shift;
//      
register unsigned char Input;
//       
register unsigned char Input_Trigger;

//    
register unsigned int RisingTime = 0;
//    
register unsigned int FallingTime = 0;

//  -    PPM 
unsigned char Sync = False;
//    FailSafe( ,    ..)
unsigned char FailSafe = True;

//    
unsigned char IN_CH_NUM;
//   
register unsigned char TX_CH = 0;
//   PPM
unsigned char PacketReceived = False;

//   (  )
unsigned char ThrCH = 2;
//    
unsigned char ThrFsCH = TimerClockPerSec(0.001);

//   
unsigned int iCH[ChannelsNum];
//   
unsigned int pCH[ChannelsNum];

// *****************************************************************************
// ***             ***************
// *****************************************************************************
#define OutPinSet(n) OUTPORT |= (1 << FCShift) << n

// *****************************************************************************
// ***             *******************
// *****************************************************************************
#define OutPinClr(n) OUTPORT &= ~((1 << FCShift) << n)

// *****************************************************************************
// ***     -        ******
// *****************************************************************************
interrupt [TIM1_COMPA] void timer1_compa_isr(void)
{
    //    0- ,   
    if(OUT_CH_NUM) OutPinClr(OUT_CH_NUM - 1);

    //     
    if(OUT_CH_NUM < TX_CH)
    {
        //   ,    
        //  75 (1.5 )    .
        //        
        //       .
        if((SendPacketCount < 75) || (OUT_CH_NUM == ThrCH)) OutPinSet(OUT_CH_NUM);
        //   
        OCR1A = pCH[OUT_CH_NUM];
        //       
        ChannelsTime += pCH[OUT_CH_NUM];
        //   
        OUT_CH_NUM++;
    }
    else
    {
        //   
        OCR1A = TimerClockPerSec(0.020) - ChannelsTime;
        //   
        ChannelsTime = 0;
        //   
        OUT_CH_NUM = 0;
        //    
        if(SendPacketCount < 0xFF) SendPacketCount++;
    }
}

// *****************************************************************************
// ***     -        *******
// *****************************************************************************
interrupt [TIM0_OVF] void timer0_ovf_isr(void)
{
    //   .
    Timer0_ovf++;
}
//     
#define Timer0_Reset() TCNT0 = 0; Timer0_ovf = 0
//      16- 
//     -     .
//   c 
#define Timer0_GetValue(p) TCCR0B=0x00; *(p) = ((unsigned int)Timer0_ovf << 8) | TCNT0; TCCR0B=0x03

// *****************************************************************************
// ***         *************************
// *****************************************************************************
void ShiftDetect(void)
{
    //    
    unsigned int Time;

    //     
    Input = INPUT;
    //     
    Input_Trigger = Input;
    //     
    Timer0_Reset();
    //        
    while((FallingTime < RetryCount) && (RisingTime < RetryCount))
    {
        //  
        while(Input == Input_Trigger) Input = INPUT;
        //   
        Input_Trigger = Input;
        //   
        Timer0_GetValue(&Time);
        //     
        Timer0_Reset();
        //  . ..     0.3 
        //      0.1 
        if(Time < TimerClockPerSec(0.0001)) continue;
        //  Input ,      
        if(Input)
        {
            //     
            if(Time > TimerClockPerSec(0.003)) RisingTime++;
            //          -  
            if((Time < TimerClockPerSec(0.0005)) && RisingTime) RisingTime = 0;
        }
        else
        {
            //     
            if(Time > TimerClockPerSec(0.003)) FallingTime++;
            //          -  
            if((Time < TimerClockPerSec(0.0005)) && FallingTime) FallingTime = 0;
        }

        //      
        if(FallingTime && RisingTime)
        {
            //    , .. 
            //      .
            RisingTime  = 0;
            FallingTime = 0;
        }
    }
    //    
    if(RisingTime) Shift = 1;
    else           Shift = 0;
}

// *****************************************************************************
// ***          ***********************
// *****************************************************************************
void NumberOfChannelsDetect(void)
{
    //    
    i = 0;
    //  
    Sync = False;
    //    n-      ,
    //        .
    while(i < RetryCount)
    {
        //   
        Input = INPUT;
        //   
        if(Input != Input_Trigger)
        {
            //      
            Input_Trigger = Input;
            //     
            //   
            if(Input == Shift)
            {
                //     
                Timer0_GetValue(&RisingTime);
            }
            //   
            else
            {
                //     
                Timer0_GetValue(&FallingTime);
                //  . ..     0.3 
                //      0.1 
                if(FallingTime - RisingTime < TimerClockPerSec(0.0001)) continue;

                //      /
                Timer0_Reset();

                //          0.5 
                if(FallingTime - RisingTime > TimerClockPerSec(0.0005))
                {
                    //  
                    Sync = False;
                    //  
                    continue;
                }

                // ..     ,  
                //        
                //  
                //     3.2  -   
                if(FallingTime > TimerClockPerSec(0.0032))
                {
                    //    
                    if(Sync == False)
                    {
                        //  
                        Sync = True;
                    }
                    //         
                    else if(TX_CH != IN_CH_NUM)
                    {
                        //    
                        TX_CH = IN_CH_NUM;
                        //   
                        i = 0;
                    }
                    //    
                    else i++;

                    //      
                    IN_CH_NUM = 0;
                    //  
                    continue;
                }

                //    ,     
                if(Sync == False) continue;

                //     ,      
                //  0.7    2.3  -   .
                //  .
                if((IN_CH_NUM >= ChannelsNum) ||
                   (FallingTime < TimerClockPerSec(0.0007)) ||
                   (FallingTime > TimerClockPerSec(0.0023)))
                {
                    //   
                    Sync = False;
                }

                //   
                IN_CH_NUM++;
            }
        }
    }
}

// *****************************************************************************
// ***    PPM    **********************************************
// *****************************************************************************
//        
//   
void PPM_Decoder(void)
{
    //   
    Input = INPUT;
    //   
    if(Input != Input_Trigger)
    {
        //      
        Input_Trigger = Input;
        //     
        //   
        if(Input == Shift)
        {
            //     
            Timer0_GetValue(&RisingTime);
        }
        //   
        else
        {
            //     
            Timer0_GetValue(&FallingTime);
            //  . ..     0.3 
            //      0.1 
            if(FallingTime - RisingTime < TimerClockPerSec(0.0001)) return;

            //      /
            Timer0_Reset();

            //          0.5 
            if(FallingTime - RisingTime > TimerClockPerSec(0.0005))
            {
                //  FailSafe
                FailSafe = True;
                //   
                Sync = False;
                // 
                return;
            }

            // ..     ,  
            //        
            //  
            //     3  -   
            if(FallingTime > TimerClockPerSec(0.003))
            {
                //  FailSafe      
                // PPM  -  FailSafe
                if((FailSafe == True) && (Sync == True)) FailSafe = False;
                //        -  FailSafe
                if(IN_CH_NUM != TX_CH) FailSafe = True;
                //  FailSafe  -   
                if(FailSafe == False) PacketReceived = True;
                //  
                Sync = True;
                //     
                IN_CH_NUM = 0;
                // 
                return;
            }

            //    ,     
            if(Sync == False) return;

            //     ,      
            //  0.7    2.3  -   .
            //  FailSafe.
            if((IN_CH_NUM >= ChannelsNum) ||
               (FallingTime < TimerClockPerSec(0.0007)) ||
               (FallingTime > TimerClockPerSec(0.0023)))
            {
                //  FailSafe
                FailSafe = True;
                //   
                Sync = False;
                //    
                PacketReceived = False;
                // 
                return;
            }

            //   
            iCH[IN_CH_NUM] = FallingTime;
            //     
            IN_CH_NUM++;
        }
    }
}

// *****************************************************************************
// ***       *****************************************************
// *****************************************************************************
void main(void)
{
    // Declare your local variables here
    // Crystal Oscillator division factor: 1
#pragma optsize-
    CLKPR=0x80;
    CLKPR=0x00;
#ifdef _OPTIMIZE_SIZE_
#pragma optsize+
#endif

    // Input/Output Ports initialization
    // Port A initialization
    // Func2=In Func1=In Func0=In
    // State2=P State1=P State0=P
    PORTA=0x07;
    DDRA=0x00;

    // Port B initialization
    // Func7=Out Func6=Out Func5=Out Func4=Out Func3=Out Func2=Out Func1=Out Func0=Out
    // State7=0 State6=0 State5=0 State4=0 State3=0 State2=0 State1=0 State0=0
    PORTB=0x00;
    DDRB=0xFF;

    // Port D initialization
    // Func6=Out Func5=Out Func4=Out Func3=Out Func2=In Func1=In Func0=In
    // State6=0 State5=0 State4=0 State3=0 State2=P State1=P State0=P
    PORTD=0x07;
    DDRD=0x78;

    // Timer/Counter 0 initialization
    // Clock source: System Clock
    // Clock value: 125,000 kHz
    // Mode: Normal top=FFh
    // OC0A output: Disconnected
    // OC0B output: Disconnected
    TCCR0A=0x00;
    TCCR0B=0x03;
    TCNT0=0x00;
    OCR0A=0x00;
    OCR0B=0x00;

    // Timer/Counter 1 initialization
    // Clock source: System Clock
    // Clock value: 125,000 kHz
    // Mode: Normal top=FFFFh
    // OC1A output: Discon.
    // OC1B output: Discon.
    // Noise Canceler: Off
    // Input Capture on Falling Edge
    // Timer 1 Overflow Interrupt: Off
    // Input Capture Interrupt: Off
    // Compare A Match Interrupt: Off
    // Compare B Match Interrupt: Off
    TCCR1A=0x00;
    TCCR1B=0x0B;
    TCNT1H=0x00;
    TCNT1L=0x00;
    ICR1H=0x00;
    ICR1L=0x00;
    OCR1AH=0x00;
    OCR1AL=0x00;
    OCR1BH=0x00;
    OCR1BL=0x00;

    // External Interrupt(s) initialization
    // INT0: Off
    // INT1: Off
    // Interrupt on any change on pins PCINT0-7: Off
    GIMSK=0x00;
    MCUCR=0x00;

    // Timer(s)/Counter(s) Interrupt(s) initialization
    TIMSK=0x02;

    // Universal Serial Interface initialization
    // Mode: Disabled
    // Clock source: Register & Counter=no clk.
    // USI Counter Overflow Interrupt: Off
    USICR=0x00;

    // Analog Comparator initialization
    // Analog Comparator: Off
    // Analog Comparator Input Capture by Timer/Counter 1: Off
    ACSR=0x80;

    // Global enable interrupts
#asm("sei")

    // ***     -      *********
    GreenLED(On);
    delay_ms(500);
    RedLED(On);
    delay_ms(500);
    GreenLED(Off);
    delay_ms(500);
    RedLED(Off);
    delay_ms(500);

    // ***          **********************
    //   
    RedLED(On);
    //   
    ShiftDetect();
    //   
    GreenLED(On);
    //     
    NumberOfChannelsDetect();
    //   
    RedLED(Off);
    // 
    delay_ms(250);

    //     
    while(PacketReceived == False) PPM_Decoder();
    //  
    PacketReceived = False;
    //  FailSafe
    FailSafe = True;
    //   
    Sync = False;
    //   
    for(i=0; i < TX_CH; i++)
    {
        //     
        pCH[i] = iCH[i];
    }
    //    
    ThrFsCH = pCH[ThrCH];

    //   
    GreenLED(Off);
    // 
    delay_ms(500);

    //   ,   
    #asm("cli")
    TIMSK |= 0x40;
    TCNT1 = 0;
    OCR1A = 0x100;
    #asm("sei")

    while(1)
    {
        //  PPM 
        PPM_Decoder();

        // ***   PPM    ***************************************************

        //     PPM
        if(PacketReceived == True)
        {
            //   -  PPM   
            PacketReceived = False;
            //   
            for(i=0; i < TX_CH; i++)
            {
                //     
                //   ,  ,  
                //   
                #asm("cli")
                pCH[i] = iCH[i];
                #asm("sei")
            }
            //      
            if(ThrFsCH > pCH[ThrCH]) ThrFsCH = pCH[ThrCH];
            //    
            SendPacketCount = 0;
        }

        //    n   -   FailSafe
        if((SendPacketCount > 3) && (FailSafe == False))
        {
            //  FailSafe
            FailSafe = True;
            //   
            Sync = False;
        }

        //   FailSafe
        if(FailSafe == True)
        {
            //   FailSafe   
            RedLED(On);
            GreenLED(Off);
            //  FailSave  15  (0,3 ) -  
            if(SendPacketCount > 15) pCH[ThrCH] = ThrFsCH;
        }
        else
        {
            //    FailSafe
            RedLED(Off);
            GreenLED(On);
        }
    };
}
